/******************************************************************************
* Include files
******************************************************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include "nmea.h"
#include "usart.h"

/******************************************************************************
* Macros
******************************************************************************/

/******************************************************************************
* Constants
******************************************************************************/


/******************************************************************************
* Global Variables
******************************************************************************/
static nmeaSentence_t nmeaSentence;

static enum nmeaStates_t
{
  NMEA_SOS,
  NMEA_SENTENCE,
  NMEA_CHECKSUM1,
  NMEA_CHECKSUM2,
  NMEA_EOS1,
  NMEA_EOS2
} nmeaState;

static nmeaPosition_t nmeaCurPos;
static uint32_t deg,min;

/******************************************************************************
* Local Function Declarations
******************************************************************************/
static void nmeaInitState(void);
static void nmeaProcessSentence(char *nmea);
static void nmeaReadByte(uint8_t b);

/******************************************************************************
* Local Function Definitions
******************************************************************************/
/******************************************************************************
* Function:      static void nmeaProcessSentence(char *nmea)
* Description:   Initialize NMEA state machine and struct
* Input:         
* Output:        nmeaState and nmeaSentence global variables
* Notes:
******************************************************************************/
static void nmeaInitState(void)
{
  nmeaState = NMEA_SOS;
  nmeaSentence.CharCount = 0;
  nmeaSentence.CheckSum = 0;
  nmeaSentence.ReceivedCheckSum = 0;
}

/******************************************************************************
* Function:      static void nmeaProcessSentence(char *nmea)
* Description:   Processes a complete NMEA sentence if it's RMC type
* Input:         NMEA string
* Output:        Global nmeaCurPos variable
* Notes:         
******************************************************************************/
static void nmeaProcessSentence(char *nmea)
{
	char *t, *u, str[10];
	

	t = nmea;
	u = t;
	
	while(*u != ',' && *u != 0)u++;
	if(*u == ',')*u++ = 0;

	// GPRMC
	if(!strcmp(t, "GPRMC"))
	{
		t = u;                                          // Time of fix
		while(*u != ',' && *u != 0)u++;
		if(*u == ',')*u++ = 0;
		if(*t == 0)return;
		strncpy(nmeaCurPos.Tof2, t, 6);
		//Hour to sec
		strncpy(str, nmeaCurPos.Tof2, 2);
		nmeaCurPos.Tof = atoi(str) * 3600;
		// Minute to sec
		strncpy(str, nmeaCurPos.Tof2 + 2, 2);
		nmeaCurPos.Tof += atoi(str) * 60;
		// Sec
		strncpy(str, nmeaCurPos.Tof2 + 4, 2);
		nmeaCurPos.Tof += atoi(str);

		t = u;                                          // Status
		while(*u != ',' && *u != 0)u++;
		if(*u == ',')*u++ = 0;
		if(*t == 'A')
		{
			nmeaCurPos.Stat = 1;
		}
		else
		{
			nmeaCurPos.Stat = 0;
		}

		t = u;                                          // Latitude
		while(*u != ',' && *u != 0)u++;
		if(*u == ',')*u++ = 0;
		deg = atoi(strncpy(str,t++,2));
		min = atoi(strncpy(str,++t,2))*100;
		t += 3;
		min += atoi(strncpy(str,t,2));
		nmeaCurPos.Lat= deg*1000000UL +min * 10000 / 60;
		
		t = u;                                          // Latitude direction
		while(*u != ',' && *u != 0)u++;
		if(*u == ',')*u++ = 0;
		if(*t == 'S')nmeaCurPos.Lat = -nmeaCurPos.Lat;
		
		t = u;                                          // Longitude
		while(*u != ',' && *u != 0)u++;
		deg = atoi(strncpy(str,t++,2));
		min = atoi(strncpy(str,++t,2))*100;
		t += 3;
		min += atoi(strncpy(str,t,2));
		nmeaCurPos.Lon= deg*1000000UL +min * 10000 / 60;
		
		t = u;                                          // Longitude direction
		while(*u != ',' && *u != 0)u++;
		if(*u == ',')*u++ = 0;
		if(*t == 'W')nmeaCurPos.Lon = -nmeaCurPos.Lon;
	}
	return;
}

/******************************************************************************
* Function:      static void nmeaReadByte(uint8_t b)
* Description:   Processes the received byte assuming NMEA sentences
* Input:         One received byte from teh buffer
* Output:
* Notes:         Calls nmeaProcessSentence when a well-formed NMEA sentence
*                with correct checksum is found.
******************************************************************************/
static void nmeaReadByte(uint8_t b)
{

  switch(nmeaState)
  {
    case NMEA_SOS :
      if(b == '$')
	  {
        nmeaState = NMEA_SENTENCE;
      }
      break;
    case NMEA_SENTENCE :
      if(b == '*')
        nmeaState = NMEA_CHECKSUM1;
      else if(b == 0x0D)
        nmeaState = NMEA_EOS2;
      else
	  {
        nmeaSentence.Sentence[nmeaSentence.CharCount++] = b;
        nmeaSentence.CheckSum ^= b;
      }
      break;
    case NMEA_CHECKSUM1 :
      if(b >= '0' && b<= '9')
	  { 
        nmeaSentence.ReceivedCheckSum = 16 * (b - '0');
        nmeaState = NMEA_CHECKSUM2;
      }
      else if(b >= 'A' && b <= 'F'){
        nmeaSentence.ReceivedCheckSum = 16 * ((b - 'A') + 10);
        nmeaState = NMEA_CHECKSUM2;
      }
      else
        nmeaInitState();
      break;  
    case NMEA_CHECKSUM2 :
      if(b >= '0' && b<= '9'){ 
        nmeaSentence.ReceivedCheckSum += (b - '0');
        if(nmeaSentence.ReceivedCheckSum == nmeaSentence.CheckSum)
          nmeaState = NMEA_EOS1;
        else 
          nmeaInitState();
      }
      else if(b >= 'A' && b <= 'F'){
        nmeaSentence.ReceivedCheckSum += (b - 'A') + 10;
        if(nmeaSentence.ReceivedCheckSum == nmeaSentence.CheckSum)
          nmeaState = NMEA_EOS1;
        else
          nmeaInitState();
      }
      else
        nmeaInitState();
      break;  
    case NMEA_EOS1 :
      if(b == 0x0D)
        nmeaState = NMEA_EOS2;
      else
        nmeaInitState();
      break;
    case NMEA_EOS2 :
      if(b == 0x0A){
        nmeaSentence.Sentence[nmeaSentence.CharCount] = 0;
        nmeaProcessSentence(nmeaSentence.Sentence);
        nmeaInitState();
      }
      else
        nmeaInitState();
      break;
  }
}

/******************************************************************************
* Global Function Definitions
******************************************************************************/
/******************************************************************************
* Function:      void nmeaInit(void)
* Description:   Initialize NMEA structs and variables
* Input:         
* Output:
* Notes:
******************************************************************************/
void nmeaInit(void)
{
	uint8_t i;
	nmeaInitState();
	nmeaCurPos.Lon = 0;
	nmeaCurPos.Lat = 0;
	nmeaCurPos.Alt = 0;
	nmeaCurPos.Vel = 0;
	nmeaCurPos.Hdg = 0;
	nmeaCurPos.Svs = 0;
	nmeaCurPos.Temp = 0;
	nmeaCurPos.Tof = 0;
	for(i=0;i<10;i++) nmeaCurPos.Tof2[i] = 0;
}

/******************************************************************************
* Function:      void nmeaGetCurrentPosition(nmeaPosition_t *d)
* Description:   Interface to the current GPS data
* Input:
* Output:        Actual GPS data
* Notes:
******************************************************************************/
void nmeaGetCurrentPosition(nmeaPosition_t *d)
{
  uint8_t i;
  
  d->Lon = nmeaCurPos.Lon;
  d->Lat = nmeaCurPos.Lat;
  d->Alt = nmeaCurPos.Alt;
  d->Vel = nmeaCurPos.Vel;
  d->Hdg = nmeaCurPos.Hdg;
  d->Svs = nmeaCurPos.Svs;
  d->Temp = nmeaCurPos.Temp;
  d->Tof = nmeaCurPos.Tof;
  for(i=0;i<10;i++) d->Tof2[i] = nmeaCurPos.Tof2[i];
}

/******************************************************************************
* Function:      void nmeaProcessReceived(void)
* Description:   Main function to precess the received NMEA sentences
* Input:
* Output:
* Notes:         Must be called periodically in the main program
******************************************************************************/
void nmeaProcessReceived(void)
{
  uint8_t rx_data;
  
  while(USART0_RxCount())
  {
    rx_data = USART0_RecieveByte();
    nmeaReadByte(rx_data);
  }
}

